from typing import cast

from .utils.typing import Frames


def correct_displaced_volume(frames: Frames, displaced_fraction: float) -> Frames:
    """Correct for displaced volume of solvent by multiplying signal by retained fraction.

    Correct for displaced volume of solvent by multiplying signal by the retained
    fraction, as detailed in section 3.xviii and appendix B of 'The modular small-angle
    X-ray scattering data correction sequence'
    [https://doi.org/10.1107/S1600576717015096].

    Args:
        frames:  A stack of frames to be corrected.
        displaced_fraction: The fraction of solvent displaced by the analyte.

    Returns:
        The corrected stack of frames.
    """
    if displaced_fraction < 0.0 or displaced_fraction > 1.0:
        raise ValueError("Displaced Fraction must be in interval [0, 1].")

    return cast(Frames, frames * (1.0 - displaced_fraction))



from math import prod

from .utils.typing import Frames


def average_all_frames(frames: Frames) -> Frames:
    """Average all frames over the leading axis.

    Args:
        frames: A stack of frames to be averaged.

    Returns:
        A frame containing the average pixel values of all frames in the stack.
    """
    return frames.reshape(
        [frames.size // prod(frames.shape[-2:]), *frames.shape[-2:]]
    ).sum(axis=-3)



from numpy import expand_dims, sum

from .utils.typing import Frames


def normalize_transmitted_flux(frames: Frames) -> Frames:
    """Normalize for incident flux and transmissibility by scaling photon counts.

    Normalize for incident flux and transmissibility by scaling photon counts with
    respect to the total observed flux, as detailed in section 4 of 'The modular small-
    angle X-ray scattering data correction sequence'
    [https://doi.org/10.1107/S1600576717015096].

    Args:
        frames: A stack of frames to be normalized.

    Returns:
        The normalized stack of frames.
    """
    frame_flux = expand_dims(sum(frames, axis=(-1, -2)), (-1, -2))
    return frames / frame_flux



from typing import Any, Literal, Tuple, Union, cast

from numpy import atleast_1d, complexfloating, dtype, expand_dims, floating, ndarray
from scipy.special import lambertw

from .utils.typing import Frames, NumFrames


def correct_deadtime(
    frames: Frames,
    count_times: ndarray[Tuple[Union[NumFrames, Literal[1]]], dtype[floating]],
    minimum_pulse_separation: float,
    minimum_arrival_separation: float,
) -> Frames:
    """Correct for detector deadtime by scaling counts to account for overlapping events.

    Correct for detector deadtime by iteratively solving for the number of incident
    photons required to produce the observed value in a given time period subject to
    detector characteristics, as detailed in section 3.3.4 of 'Everything SAXS: small-
    angle scattering pattern collection and correction'
    [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        frames: A stack of frames to be corrected.
        count_times: The period over which photons are counted for each frame.
        minimum_pulse_separation: The minimum time difference required between a prior
            pulse and the current pulse for the current pulse to be recorded correctly.
        minimum_arrival_separation: The minimum time difference required between the
            current pulse and a subsequent pulse for the current pulse to be recorded
            correctly.

    Returns:
        The corrected stack of frames.
    """
    if (count_times <= 0).any():
        raise ValueError("Count times must be positive.")
    if minimum_pulse_separation < 0:
        raise ValueError("Minimum Pulse Separation must be non-negative.")
    if minimum_arrival_separation < 0:
        raise ValueError("Minimum Arrival Separation must be non-negative.")

    if minimum_pulse_separation == 0 and minimum_arrival_separation == 0:
        return frames

    deadtime_proportion = expand_dims(
        atleast_1d(
            (minimum_pulse_separation + minimum_arrival_separation) / count_times
        ),
        (1, 2),
    )
    return (
        -cast(
            ndarray[Any, dtype[complexfloating]],
            lambertw(-deadtime_proportion * frames),
        ).real
        / deadtime_proportion
    )



from typing import Tuple, TypeVar, cast

from numpy import cos, dtype, power

from .utils.geometry import scattering_angles
from .utils.typing import Frames

FrameDType = TypeVar("FrameDType", bound=dtype)
FramesShape = TypeVar("FramesShape", bound=Tuple[int, int, int])


def correct_solid_angle(
    frames: Frames,
    beam_center: Tuple[float, float],
    pixel_sizes: Tuple[float, float],
    distance: float,
) -> Frames:
    """Corrects for the solid angle by scaling by the inverse of area subtended by a pixel.

    Corrects for the solid angle by scaling by the inverse of area subtended by each
    pixel, as detailed in section 3.4.6 of 'Everything SAXS: small-angle scattering
    pattern collection and correction' [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        frames: A stack of frames to be corrected.
        beam_center: The center position of the beam in pixels.
        pixel_sizes: The real space size of a detector pixel.
        distance: The distance between the detector and the sample head.

    Returns:
        The corrected stack of frames.
    """
    correction = power(
        cos(
            scattering_angles(
                cast(Tuple[int, int], frames.shape[-2:]),
                beam_center,
                pixel_sizes,
                distance,
            )
        ),
        3,
    )
    return frames / correction



from typing import Any

from numpy import bool_, broadcast_to, dtype, ndarray
from numpy.ma import MaskedArray, masked_where

from .utils.typing import FrameDType, StackShape


def mask_frames(
    frames: ndarray[StackShape, FrameDType], mask: ndarray[Any, dtype[bool_]]
) -> MaskedArray[StackShape, FrameDType]:
    """Replaces masked elemenets of frames in a stack with zero.

    Args:
        frames: A stack of frames to be masked.
        mask: The boolean mask to apply to each frame.

    Returns:
        A stack of frames where pixels.
    """
    return masked_where(broadcast_to(mask, frames.shape), frames)



from .utils.typing import Frame, Frames


def subtract_background(foreground_frames: Frames, background_frame: Frame) -> Frames:
    """Subtract a background frame from a sequence of foreground frames.

    Subtract a background frame from a sequence of foreground frames, as detailed in
    section 3.4.6 of 'Everything SAXS: small-angle scattering pattern collection and
    correction' [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        foreground_frames: A sequence of foreground frames to be corrected.
        background_frame: The background which is to be corrected for.

    Returns:
        A sequence of corrected frames.
    """
    return foreground_frames - background_frame



from typing import List, Tuple

from numpy import arctan, arctan2, dtype, floating, hypot, linspace, meshgrid, ndarray


def _relative_position_meshgrid(
    frame_shape: Tuple[int, int],
    beam_center: Tuple[float, float],
    pixel_sizes: Tuple[float, float],
) -> List[ndarray]:
    return meshgrid(
        linspace(
            (0.5 - beam_center[1]) * pixel_sizes[1],
            (frame_shape[1] - 0.5 - beam_center[1]) * pixel_sizes[1],
            frame_shape[1],
        ),
        linspace(
            (0.5 - beam_center[0]) * pixel_sizes[0],
            (frame_shape[0] - 0.5 - beam_center[0]) * pixel_sizes[0],
            frame_shape[0],
        ),
    )


def scattering_angles(
    frame_shape: Tuple[int, int],
    beam_center: Tuple[float, float],
    pixel_sizes: Tuple[float, float],
    distance: float,
) -> ndarray[Tuple[int, int], dtype[floating]]:
    """Computes the angles of pixels from the sample for a given geometry.

    Args:
        frame_shape: The shape of a frame.
        beam_center: The center position of the beam in pixels.
        pixel_sizes: The real space size of a detector pixel.
        distance: The distance between the detector and the sample.

    Returns:
        An array of pixel angles from the sample.
    """
    yy, xx = _relative_position_meshgrid(frame_shape, beam_center, pixel_sizes)
    return arctan(hypot(xx, yy) / distance)


def azimuthal_angles(
    frame_shape: Tuple[int, int],
    beam_center: Tuple[float, float],
    pixel_sizes: Tuple[float, float],
) -> ndarray[Tuple[int, int], dtype[floating]]:
    """Computes the azimuthal angles of pixels from the beam center.

    Args:
        frame_shape: The shape of the frame.
        beam_center: The center position of the beam in pixels.

    Returns:
        An array of pixel azimuthal angles from the beam center.
    """
    yy, xx = _relative_position_meshgrid(frame_shape, beam_center, pixel_sizes)
    return arctan2(xx, yy)



from . import utils
from ._version_git import __version__
from .angular_efficiency import correct_angular_efficiency
from .background_subtraction import subtract_background
from .dark_current import correct_dark_current
from .deadtime import correct_deadtime
from .displaced_volume import correct_displaced_volume
from .flatfield import correct_flatfield
from .flux_and_transmission import normalize_transmitted_flux
from .frame_average import average_all_frames
from .frame_time import normalize_frame_time
from .masking import mask_frames
from .polarization import correct_polarization
from .self_absorption import correct_self_absorption
from .solid_angle import correct_solid_angle
from .thickness import normalize_thickness

__all__ = [
    "__version__",
    "utils",
    "mask_frames",
    "correct_deadtime",
    "correct_dark_current",
    "normalize_frame_time",
    "normalize_transmitted_flux",
    "correct_self_absorption",
    "average_all_frames",
    "subtract_background",
    "correct_flatfield",
    "correct_angular_efficiency",
    "correct_solid_angle",
    "correct_polarization",
    "normalize_thickness",
    "correct_displaced_volume",
]



from typing import Tuple, TypeVar

from numpy import dtype, ndarray

#: The underlying data type of a frame.
FrameDType = TypeVar("FrameDType", bound=dtype)
#: The number of frames in a stack of frames.
NumFrames = TypeVar("NumFrames", bound=int)
#: The number of pixels spanning the width of a frame.
FrameWidth = TypeVar("FrameWidth", bound=int)
#: The number of pixels spanning the height of a frame.
FrameHeight = TypeVar("FrameHeight", bound=int)

#: The shape of a frame; Comprising of a width and height.
FrameShape = Tuple[FrameWidth, FrameHeight]
#: The shape of a stack of frames; Comprising a count, width and height.
StackShape = Tuple[NumFrames, FrameWidth, FrameHeight]

#: A frame; Comprising a shape and a data type
Frame = ndarray[FrameShape, FrameDType]
#: A stack of frames; Comprising a shape and a data type
Frames = ndarray[StackShape, FrameDType]



from math import exp
from typing import Tuple, cast

from numpy import (
    cos,
    divide,
    dtype,
    floating,
    log,
    logical_and,
    ndarray,
    ones_like,
    power,
)

from .utils.geometry import scattering_angles
from .utils.typing import Frames


def correct_self_absorption(
    frames: Frames,
    beam_center: Tuple[float, float],
    pixel_sizes: Tuple[float, float],
    distance: float,
    absorption_coefficient: float,
    thickness: float,
) -> Frames:
    """Correct for transmission loss due to differences in observation angle.

    Correct for transmission loss due to differences in observation angle, as detailed
    in section 3.4.7 of 'Everything SAXS: small-angle scattering pattern collection and
    correction' [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        frames: A stack of frames to be corrected.
        beam_center: The center position of the beam in pixels.
        pixel_sizes: The real space size of a detector pixel.
        distance: The distance between the detector and the sample.
        absorption_coefficient: The coefficient of absorption for a given
            material at a given photon energy.
        thickness: The thickness of the detector head material.

    Returns:
        The corrected stack of frames.
    """
    if absorption_coefficient < 0.0:
        raise ValueError("absorption coefficient must non-negative.")
    if thickness < 0.0:
        raise ValueError("Thickness must be non-negative.")

    angles = scattering_angles(
        cast(Tuple[int, int], frames.shape[-2:]), beam_center, pixel_sizes, distance
    )
    transmissibility = exp(-absorption_coefficient * thickness)
    secangle: ndarray[Tuple[int, int], dtype[floating]] = 1 / cos(angles)
    correction_factors: ndarray[Tuple[int, int], dtype[floating]] = divide(
        1 - power(transmissibility, secangle - 1),
        log(transmissibility) * (1 - secangle),
        out=ones_like(secangle),
        where=logical_and(secangle != 1.0, transmissibility != 1.0),
    )

    return frames * correction_factors



from typing import Tuple, cast

from numpy import cos, sin, square

from .utils.geometry import azimuthal_angles, scattering_angles
from .utils.typing import Frames


def correct_polarization(
    frames: Frames,
    beam_center: Tuple[float, float],
    pixel_sizes: Tuple[float, float],
    distance: float,
    horizontal_poarization: float = 0.5,
) -> Frames:
    """Corrects for the effect of polarization of the incident beam.

    Corrects for the effect of polarization of the incident beam, as detailed in
    section 3.4.1 of 'Everything SAXS: small-angle scattering pattern collection and
    correction' [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        frames: A stack of frames to be corrected.
        beam_center: The center position of the beam in pixels.
        pixel_sizes: The real space size of a detector pixel.
        distance: The distance between the detector and the sample.
        horizontal_poarization: The fraction of incident radiation polarized in the
            horizontal plane, where 0.5 signifies an unpolarized source. Defaults to
            0.5.

    Returns:
        The corrected stack of frames.
    """
    if horizontal_poarization < 0.0 or horizontal_poarization > 1.0:
        raise ValueError("Horizontal Polarization must be within the interval [0, 1].")

    scattering = scattering_angles(
        cast(tuple[int, int], frames.shape[-2:]), beam_center, pixel_sizes, distance
    )
    azimuths = azimuthal_angles(
        cast(tuple[int, int], frames.shape[-2:]), beam_center, pixel_sizes
    )
    correction_factors = horizontal_poarization * (
        1.0 - square(sin(azimuths) * sin(scattering))
    ) + (1.0 - horizontal_poarization) * (1.0 - square(cos(azimuths) * sin(scattering)))
    return frames * correction_factors



import click

from . import __version__


@click.group(invoke_without_command=True)
@click.version_option(version=__version__, message="%(version)s")
@click.pass_context
def main(ctx: click.Context) -> None:
    """The main command line interface entry point.

    Args:
        ctx: The click context.
    """
    if ctx.invoked_subcommand is None:
        click.echo(main.get_help(ctx))



from typing import cast

from .utils.typing import Frames


def normalize_thickness(
    frames: Frames,
    sample_thickness: float,
) -> Frames:
    """Normailizes pixel intensities by dividing by the sample thickness.

    Normailizes pixel intensities by dividing by the sample thickness, as detailed in
    section 3.4.3 of 'Everything SAXS: small-angle scattering pattern collection and
    correction' [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        frames: A stack of frames to be corrected.
        sample_thickness: The thickness of the exposed sample.

    Returns:
        The normalized stack of frames.
    """
    if sample_thickness <= 0:
        raise ValueError("Sample Thickness must be positive.")

    return cast(Frames, frames / sample_thickness)



from typing import Literal, Tuple, Union

from numpy import atleast_1d, dtype, expand_dims, floating, ndarray

from .utils.typing import Frames, NumFrames


def correct_dark_current(
    frames: Frames,
    count_times: ndarray[Tuple[Union[NumFrames, Literal[1]]], dtype[floating]],
    base_dark_current: float,
    temporal_dark_current: float,
    flux_dependant_dark_current: float,
) -> Frames:
    """Correct by subtracting base, temporal and flux-dependant dark currents.

    Correct for incident dark current by subtracting a baselike, time dependant and a
    flux dependant count rate from frames, as detailed in section 3.3.6 of 'Everything
    SAXS: small-angle scattering pattern collection and correction'
    [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        frames: A stack of frames to be corrected.
        count_times: The period over which photons are counted for each frame.
        base_dark_current: The dark current flux, irrespective of time.
        temporal_dark_current: The dark current flux, as a factor of time.
        flux_dependant_dark_current: The dark current flux, as a factor of incident
            flux.

    Returns:
        The corrected stack of frames.
    """
    if (count_times <= 0).any():
        raise ValueError("Count times must be positive.")
    if base_dark_current < 0:
        raise ValueError("Base Dark Current must be non-negative.")
    if temporal_dark_current < 0:
        raise ValueError("Temporal Dark Current must be non-negative.")
    if flux_dependant_dark_current < 0:
        raise ValueError("Flux Dependant Dark Current must be non-negative.")

    temporal = temporal_dark_current * expand_dims(atleast_1d(count_times), (1, 2))
    flux_dependant = flux_dependant_dark_current * frames
    return frames - base_dark_current - temporal - flux_dependant



from typing import Literal, Tuple, Union

from numpy import atleast_1d, dtype, expand_dims, floating, ndarray

from .utils.typing import Frames, NumFrames


def normalize_frame_time(
    frames: Frames,
    count_times: ndarray[Tuple[Union[NumFrames, Literal[1]]], dtype[floating]],
) -> Frames:
    """Normalize for detector frame rate by scaling photon counts according to count time.

    Normalize for detector frame rate by scaling photon counts according to count time,
    as detailed in section 3.4.3 of 'Everything SAXS: small-angle scattering pattern
    collection and correction' [https://doi.org/10.1088/0953-8984/25/38/383201].

    Args:
        frames: A stack of frames to be normalized.
        count_times: The period over which photons are counted for each frame.

    Returns:
        The normalized stack of frames.
    """
    if (count_times <= 0).any():
        raise ValueError("Count times must be positive.")

    times = expand_dims(atleast_1d(count_times), (1, 2))
    return frames / times



from numpy import dtype, floating, ndarray

from .utils.typing import Frames, FrameShape


def correct_flatfield(
    frames: Frames, flatfield: ndarray[FrameShape, dtype[floating]]
) -> Frames:
    """Apply multiplicative flatfield correction, to correct for inter-pixel sensitivity.

    Apply multiplicative flatfield correction, to correct for inter-pixel sensitivity,
    as described in section 3.xii of 'The modular small-angle X-ray scattering data
    correction sequence' [https://doi.org/10.1107/S1600576717015096].

    Args:
        frames: A stack of frames to be corrected.
        flatfield: The multiplicative flatfield correction to be applied.

    Returns:
        The corrected stack of frames.
    """
    return frames * flatfield



from typing import Tuple, cast

from numpy import cos, exp

from .utils.geometry import scattering_angles
from .utils.typing import Frames


def correct_angular_efficiency(
    frames: Frames,
    beam_center: Tuple[float, float],
    pixel_sizes: Tuple[float, float],
    distance: float,
    absorption_coefficient: float,
    thickness: float,
) -> Frames:
    """Corrects for loss due to the angular efficiency of the detector head.

    Corrects for loss due to the angular efficiency of the detector head, as described
    in section 3.xiii and appendix C of 'The modular small-angle X-ray scattering data
    correction sequence' [https://doi.org/10.1107/S1600576717015096].

    Args:
        frames: A stack of frames to be corrected.
        beam_center: The center position of the beam in pixels.
        pixel_sizes: The real space size of a detector pixel.
        distance: The distance between the detector and the sample head.
        absorption_coefficient: The coefficient of absorption for a given material at a
            given photon energy.
        thickness: The thickness of the detector head material.

    Returns:
        The corrected stack of frames.
    """
    if absorption_coefficient <= 0.0:
        raise ValueError("Absorption Coefficient must positive.")
    if thickness <= 0.0:
        raise ValueError("Thickness must be positive.")

    absorption_efficiency = 1.0 - exp(
        -absorption_coefficient
        * thickness
        / cos(
            scattering_angles(
                cast(tuple[int, int], frames.shape[-2:]),
                beam_center,
                pixel_sizes,
                distance,
            )
        )
    )
    return frames / absorption_efficiency
